import asyncio
import logging
import sys
from enum import Enum
from typing import Callable

from asyncfix import FTag
from asyncfix.codec import Codec
from asyncfix.engine import FIXEngine
from asyncfix.journaler import DuplicateSeqNoError
from asyncfix.message import FIXMessage, MessageDirection
from asyncfix.protocol import FIXProtocolBase
from asyncfix.session import FIXSession


class ConnectionState(Enum):
    UNKNOWN = 0
    DISCONNECTED = 1
    CONNECTED = 2
    LOGGED_IN = 3
    LOGGED_OUT = 4


class FIXException(Exception):
    class FIXExceptionReason(Enum):
        NOT_CONNECTED = 0
        DECODE_ERROR = 1
        ENCODE_ERROR = 2

    def __init__(self, reason, description=None):
        super(Exception, self).__init__(description)
        self.reason = reason


class SessionWarning(Exception):
    pass


class SessionError(Exception):
    pass


class FIXConnectionHandler(object):
    def __init__(
        self,
        engine: FIXEngine,
        protocol: FIXProtocolBase,
        socket_reader: asyncio.StreamReader,
        socket_writer: asyncio.StreamWriter,
        addr=None,
        observer=None,
    ):
        self.codec = Codec(protocol)
        self.engine = engine
        self.connection_state = ConnectionState.CONNECTED
        self.session: FIXSession | None = None
        self.addr = addr
        self.observer = observer
        self.msg_buffer = b""
        self.heartbeat_period = 30.0
        self.msg_handlers = []
        self.socket_reader = socket_reader
        self.socket_writer = socket_writer
        self.heartbeat_timer_registration = None
        self.expected_heartbeat_egistration = None
        # self.reader_task = asyncio.create_task(self.handle_read())

    def address(self):
        return self.addr

    async def disconnect(self):
        await self.handle_close()

    async def _notify_message_observers(
        self,
        msg: FIXMessage,
        direction: MessageDirection,
        persist_message=True,
    ):
        if persist_message is True:
            self.engine.journaller.persist_msg(msg, self.session, direction)
        for handler in filter(
            lambda x: (x[1] is None or x[1] == direction)
            and (x[2] is None or x[2] == msg.msg_type),
            self.msg_handlers,
        ):
            await handler[0](self, msg)

    def add_message_handler(
        self,
        handler: Callable,
        direction: MessageDirection | None = None,
        msg_type=None,
    ):
        self.msg_handlers.append((handler, direction, msg_type))

    def remove_message_handler(
        self,
        handler: Callable,
        direction: MessageDirection | None = None,
        msg_type=None,
    ):
        remove = filter(
            lambda x: x[0] == handler
            and (x[1] == direction or direction is None)
            and (x[2] == msg_type or msg_type is None),
            self.msg_handlers,
        )
        for h in remove:
            self.msg_handlers.remove(h)

    def _send_heartbeat(self):
        self.send_msg(self.codec.protocol.heartbeat())

    def _expected_heartbeat(self, type, closure):
        logging.warning(
            "Expected heartbeat from peer %s" % (self.expected_heartbeat_egistration,)
        )
        self.send_msg(self.codec.protocol.test_request())

    def _handle_resend_request(self, msg: FIXMessage):
        protocol: FIXProtocolBase = self.codec.protocol
        responses = []

        begin_seq_no = msg[FTag.BeginSeqNo]
        end_seq_no = msg[FTag.EndSeqNo]
        if int(end_seq_no) == 0:
            end_seq_no = sys.maxsize
        logging.info("Received resent request from %s to %s", begin_seq_no, end_seq_no)
        replay_msgs = self.engine.journaller.recover_msgs(
            self.session, MessageDirection.OUTBOUND, begin_seq_no, end_seq_no
        )
        gap_fill_begin = int(begin_seq_no)
        gap_fill_end = int(begin_seq_no)
        for replay_msg in replay_msgs:
            msg_seq_num = int(replay_msg[FTag.MsgSeqNum])
            if replay_msg[FTag.MsgType] in protocol.session_message_types:
                gap_fill_end = msg_seq_num + 1
            else:
                if self.engine.should_resend_message(self.session, replay_msg):
                    if gap_fill_begin < gap_fill_end:
                        # we need to send a gap fill message
                        gap_fill_msg = FIXMessage(protocol.msgtype.SEQUENCERESET)
                        gap_fill_msg[FTag.GapFillFlag] = "Y"
                        gap_fill_msg[FTag.MsgSeqNum] = gap_fill_begin
                        gap_fill_msg[FTag.NewSeqNo] = str(gap_fill_end)
                        responses.append(gap_fill_msg)

                    # and then resent the replayMsg
                    del replay_msg[FTag.BeginString]
                    del replay_msg[FTag.BodyLength]
                    del replay_msg[FTag.SendingTime]
                    del replay_msg[FTag.SenderCompID]
                    del replay_msg[FTag.TargetCompID]
                    del replay_msg[FTag.CheckSum]
                    replay_msg[FTag.PossDupFlag] = "Y"
                    responses.append(replay_msg)

                    gap_fill_begin = msg_seq_num + 1
                else:
                    gap_fill_end = msg_seq_num + 1
                    responses.append(replay_msg)

        if gap_fill_begin < gap_fill_end:
            # we need to send a gap fill message
            gap_fill_msg = FIXMessage(protocol.msgtype.SEQUENCERESET)
            gap_fill_msg[FTag.GapFillFlag] = "Y"
            gap_fill_msg[FTag.MsgSeqNum] = gap_fill_begin
            gap_fill_msg[FTag.NewSeqNo] = str(gap_fill_end)
            responses.append(gap_fill_msg)

        return responses

    async def handle_read(self):
        logging.debug("handle_read")
        while True:
            try:
                msg = await self.socket_reader.read(4096)
                if not msg:
                    raise ConnectionError
                # breakpoint()

                self.msg_buffer = self.msg_buffer + msg
                while True:
                    if self.connection_state == ConnectionState.DISCONNECTED:
                        break

                    (decoded_msg, parsed_length) = self.codec.decode(self.msg_buffer)

                    if parsed_length > 0:
                        self.msg_buffer = self.msg_buffer[parsed_length:]

                    if decoded_msg is None:
                        break

                    await self.process_message(decoded_msg)
            except asyncio.CancelledError:
                break
            except ConnectionError as why:
                logging.debug("Connection has been closed %s" % (why,))
                await self.disconnect()
                return
            except Exception:
                logging.exception("handle_read failed")
                raise

    async def handle_session_message(self, msg: FIXMessage):
        return -1

    async def process_message(self, decoded_msg: FIXMessage):
        protocol: FIXProtocolBase = self.codec.protocol
        logging.debug(f"processMessage \n\t {decoded_msg}")

        begin_string = decoded_msg[FTag.BeginString]
        if begin_string != protocol.beginstring:
            logging.warning(
                "FIX BeginString is incorrect (expected: %s received: %s)",
                (protocol.beginstring, begin_string),
            )
            await self.disconnect()
            return

        msg_type = decoded_msg[FTag.MsgType]

        try:
            responses = []
            if msg_type in protocol.session_message_types:
                (recv_seq_no, responses) = await self.handle_session_message(
                    decoded_msg
                )
            else:
                recv_seq_no = decoded_msg[FTag.MsgSeqNum]

            # validate the seq number
            (seq_no_state, last_known_seq_no) = self.session.validate_recv_seq_no(
                recv_seq_no
            )

            if seq_no_state is False:
                # We should send a resend request
                logging.info(
                    "Requesting resend of messages: %s to %s" % (last_known_seq_no, 0)
                )
                responses.append(protocol.resend_request(last_known_seq_no, 0))
                # we still need to notify if we are processing Logon message
                if msg_type == protocol.msgtype.LOGON:
                    await self._notify_message_observers(
                        decoded_msg, MessageDirection.INBOUND, False
                    )
            else:
                self.session.set_recv_seq_no(recv_seq_no)
                await self._notify_message_observers(
                    decoded_msg, MessageDirection.INBOUND
                )

            for m in responses:
                await self.send_msg(m)

        except SessionWarning as sw:
            logging.warning(sw)
        except SessionError as se:
            logging.error(se)
            await self.disconnect()
        except DuplicateSeqNoError:
            try:
                if decoded_msg[FTag.PossDupFlag] == "Y":
                    logging.debug("Received duplicate message with PossDupFlag set")
            except KeyError:
                pass
            finally:
                logging.error(
                    "Failed to process message with duplicate seq no (MsgSeqNum: %s)"
                    " (and no PossDupFlag='Y') - disconnecting" % (recv_seq_no,)
                )
                await self.disconnect()

    async def handle_close(self):
        if self.connection_state != ConnectionState.DISCONNECTED:
            logging.info("Client disconnected")
            self.socket_writer.close()
            self.connection_state = ConnectionState.DISCONNECTED
            self.msg_handlers.clear()
            if self.observer is not None:
                await self.observer.notify_disconnect(self)

    async def send_msg(self, msg: FIXMessage):
        if (
            self.connection_state != ConnectionState.CONNECTED
            and self.connection_state != ConnectionState.LOGGED_IN
        ):
            logging.debug("sendMsg not connected or logged in")
            return

        encoded_msg = self.codec.encode(msg, self.session).encode("utf-8")
        self.socket_writer.write(encoded_msg)
        await self.socket_writer.drain()
        decoded_msg, junk = self.codec.decode(encoded_msg)
        logging.debug(f"send msg sending msg\n\t {decoded_msg}")

        try:
            await self._notify_message_observers(decoded_msg, MessageDirection.OUTBOUND)
        except DuplicateSeqNoError:
            logging.error(
                "We have sent a message with a duplicate seq no, failed to persist it"
                " (MsgSeqNum: %s)" % (decoded_msg[self.codec.FTag.MsgSeqNum])
            )


class FIXEndPoint(object):
    def __init__(self, engine: FIXEngine, protocol: FIXProtocolBase):
        self.engine: FIXEngine = engine
        self.protocol: FIXProtocolBase = protocol

        self.connections: list[FIXConnectionHandler] = []
        self.message_handlers = []

    def writable(self):
        return True

    def start(self, host, port, loop):
        pass

    def stop(self):
        pass

    def add_connection_listener(self, handler: Callable, filter):
        self.message_handlers.append((handler, filter))

    def remove_connection_listener(self, handler: Callable, filter):
        for s in self.message_handlers:
            if s == (handler, filter):
                self.message_handlers.remove(s)

    async def notify_disconnect(self, connection: FIXConnectionHandler):
        self.connections.remove(connection)
        for handler in filter(
            lambda x: x[1] == ConnectionState.DISCONNECTED, self.message_handlers
        ):
            await handler[0](connection)



import pickle
import unittest

import pytest

from asyncfix import FIXMessage, FTag
from asyncfix.errors import (
    DuplicatedTagError,
    FIXMessageError,
    RepeatingTagError,
    TagNotFoundError,
)
from asyncfix.message import FIXContainer


def test_init_construction():
    msg = FIXMessage(
        "AB", {11: "clordis", "1": "account", FTag.Price: 21.21, FTag.OrderQty: 2}
    )
    assert msg[11] == "clordis"
    assert msg[FTag.Account] == "account"
    assert msg[FTag.Price] == "21.21"
    assert msg[FTag.OrderQty] == "2"


def test_tag_errors():
    msg = FIXMessage("AB")
    msg["45"] = "dgd"

    with pytest.raises(TagNotFoundError) as exc:
        msg["99"]
    with pytest.raises(TagNotFoundError) as exc:
        msg.get("99")

    assert msg.get("99", 100) == 100

    with pytest.raises(DuplicatedTagError, match="tag=45 already exists"):
        msg[45] = "aaa"

    assert 45 in msg
    del msg[45]
    assert 45 not in msg

    with pytest.raises(FIXMessageError, match="Tags must be only integers"):
        msg["abs"] = "aaa"

    msg[45] = RepeatingTagError
    with pytest.raises(
        RepeatingTagError, match="tag=45 was repeated, possible undefined "
    ):
        msg[45]


def test_groups():
    msg = FIXMessage("AB")

    msg.set_group(2023, [{1: "a", 2: "b"}, FIXContainer({1: "c", 2: "d"})])

    assert msg.get_group_by_tag(2023, 1, "a") == {1: "a", 2: "b"}

    with pytest.raises(
        TagNotFoundError, match="get_group_by_tag: tag=2023 gtag=1 gvalue='z' missing"
    ):
        assert msg.get_group_by_tag(2023, 1, "z") == "1=a|2=b"

    with pytest.raises(
        DuplicatedTagError, match="group with tag='2023' already exists"
    ):
        msg.set_group(2023, [{1: "a", 2: "b"}, FIXContainer({1: "c", 2: "d"})])

    with pytest.raises(
        FIXMessageError, match="groups must be a list of FIXContext or dict"
    ):
        msg.set_group(2022, [{1: "a", 2: "b"}, 2123])

    msg.add_group(2023, {1: "e", 4: "f"})

    with pytest.raises(FIXMessageError, match="Expected FIXContext in group, got "):
        msg.add_group(2023, None)

    g = msg.get_group_by_index(2023, 0)
    assert isinstance(g, FIXContainer)
    assert g[1] == "a"
    assert g[2] == "b"

    g = msg.get_group_by_index(2023, 1)
    assert isinstance(g, FIXContainer)
    assert g[1] == "c"
    assert g[2] == "d"

    g = msg.get_group_by_index(2023, 2)
    assert isinstance(g, FIXContainer)
    assert g[1] == "e"
    assert g[4] == "f"

    with pytest.raises(
        TagNotFoundError, match="index is out of range of tag=2023 group"
    ):
        g = msg.get_group_by_index(2023, 3)

    assert msg.is_group(2023)

    with pytest.raises(
        FIXMessageError, match="You are trying to get group as simple tag"
    ):
        # Getting group as simple tag is not allowed
        g = msg[2023]


def test_init_group_construction():
    msg = FIXMessage(
        "AB",
        {
            11: "clordis",
            "1": "account",
            2023: [{1: "a", 2: "b"}, FIXContainer({1: "c", 2: "d"})],
        },
    )

    g = msg.get_group_by_index(2023, 0)
    assert isinstance(g, FIXContainer)
    assert g[1] == "a"
    assert g[2] == "b"

    g = msg.get_group_by_index(2023, 1)
    assert isinstance(g, FIXContainer)
    assert g[1] == "c"
    assert g[2] == "d"


def test_msg_construction():
    msg = FIXMessage("AB")
    msg["45"] = "dgd"
    assert msg["45"] == "dgd"
    assert "45" in msg
    assert 45 in msg

    msg.set("32", "aaaa")
    msg.set("323", "bbbb")

    rptgrp1 = FIXContainer()
    rptgrp1.set("611", "aaa")
    rptgrp1.set("612", "bbb")
    rptgrp1.set("613", "ccc")

    msg.add_group("444", rptgrp1, 0)

    rptgrp2 = FIXContainer({611: "zzz", 612: "yyy", "613": "xxx"})
    msg.add_group("444", rptgrp2, 1)

    assert "45=dgd|32=aaaa|323=bbbb|444=2=>[611=aaa|612=bbb|613=ccc,"
    " 611=zzz|612=yyy|613=xxx]" == str(msg)

    msg.add_group("444", rptgrp2, 1)

    rptgrp3 = FIXContainer()
    rptgrp3.set("611", "ggg")
    rptgrp3.set("612", "hhh")
    rptgrp3.set("613", "jjj")
    rptgrp2.add_group("445", rptgrp3, 0)
    assert "45=dgd|32=aaaa|323=bbbb|444=2=>[611=aaa|612=bbb|613=ccc,"
    " 611=zzz|612=yyy|613=xxx|445=1=>[611=ggg|612=hhh|613=jjj]]" == str(msg)

    grp = msg.get_group_by_tag("444", "612", "yyy")
    assert "611=zzz|612=yyy|613=xxx|445=1=>[611=ggg|612=hhh|613=jjj]" == str(grp)


def testPickle():
    msg = FIXMessage("AB")
    msg.set("45", "dgd")
    msg.set("32", "aaaa")
    msg.set("323", "bbbb")

    rptgrp1 = FIXContainer()
    rptgrp1.set("611", "aaa")
    rptgrp1.set("612", "bbb")
    rptgrp1.set("613", "ccc")

    msg.add_group("444", rptgrp1, 0)

    str = pickle.dumps(msg)

    msg2 = pickle.loads(str)
    assert msg == msg2


def test_dict_contains():
    msg = FIXMessage(
        "AB", {11: "clordis", "1": "account", FTag.Price: 21.21, FTag.OrderQty: 2}
    )
    assert msg[11] == "clordis"
    assert msg[FTag.Account] == "account"
    assert msg[FTag.Price] == "21.21"
    assert msg[FTag.OrderQty] == "2"

    assert {11: "clordis"} in msg
    assert {12: "clordis"} not in msg
    assert {11: "clordis", "2": "account"} not in msg
    assert {11: "clordis1"} not in msg

    msg.add_group("444", {1: "ok"})
    with pytest.raises(
        FIXMessageError,
        match="__contains__ supports only simple tags, got group at tag=444",
    ):
        assert {444: "clordis"} in msg


def test_dict_equals():
    msg = FIXMessage(
        "AB", {11: "clordis", "1": "account", FTag.Price: 21.21, FTag.OrderQty: 2}
    )
    assert msg[11] == "clordis"
    assert msg[FTag.Account] == "account"
    assert msg[FTag.Price] == "21.21"
    assert msg[FTag.OrderQty] == "2"

    assert msg != "aaldsa"

    assert {11: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} in msg
    assert {11: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} == msg
    assert {12: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} != msg
    assert {11: "clordis1", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} != msg

    # Ignoring tags
    msg[FTag.BeginString] = "FIX.4.4"
    assert {11: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} == msg
    msg[FTag.BodyLength] = "100"
    assert {11: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} == msg
    msg[FTag.CheckSum] = "100"
    assert {11: "clordis", 1: "account", FTag.Price: 21.21, FTag.OrderQty: 2} == msg
    msg[FTag.MsgType] = "1"
    assert {11: "clordis", "1": "account", FTag.Price: 21.21, FTag.OrderQty: 2} == msg

    # Other tag fails
    msg[2000] = "1"
    assert {11: "clordis", "1": "account", FTag.Price: 21.21, FTag.OrderQty: 2} != msg

    assert {
        11: "clordis",
        "1": "account",
        FTag.Price: 21.21,
        FTag.OrderQty: 2,
        2000: 1,
    } == msg

    msg.add_group(555, {1: "foo"})
    with pytest.raises(
        FIXMessageError,
        match="fix message __eq__ .* supports only simple tags, got group at tag=555",
    ):
        assert {
            11: "clordis",
            "1": "account",
            FTag.Price: 21.21,
            FTag.OrderQty: 2,
            2000: 1,
            555: [{1: "foo"}],
        } == msg



from datetime import datetime
from math import isfinite, nan

from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.errors import FIXError

from .common import FExecType, FOrdSide, FOrdStatus, FOrdType


class FIXNewOrderSingle:
    def __init__(
        self,
        clord_id: str,
        cl_ticker: str,
        side: FOrdSide | str,
        price: float,
        qty: float,
        ord_type: FOrdType | str = FOrdType.LIMIT,
        account: str | dict = "000000",
        target_price: float | None = None,
    ):
        self.clord_id = clord_id
        self.orig_clord_id = None
        self.order_id = None
        self.ticker = cl_ticker
        self.side = side
        self.price = price
        self.qty = qty
        self.leaves_qty = 0.0
        self.cum_qty = 0.0
        self.ord_type = ord_type
        self.account = account
        self.clord_id_cnt = 0
        self.status: FOrdStatus = FOrdStatus.CREATED
        self.target_price = target_price if target_price is not None else price

    def next_clord(self) -> str:
        """
        New ClOrdID for current order management
        """
        self.clord_id_cnt += 1
        return f"{self.clord_id}-{self.clord_id_cnt}"

    @staticmethod
    def current_datetime():
        """
        TransactTime field
        """
        return datetime.utcnow().strftime("%Y%m%d-%H:%M:%S.%f")[:-3]

    def new_req(self) -> FIXMessage:
        """
        Creates NewOrderSingle message
        """
        o = FIXMessage(FMsg.NEWORDERSINGLE)
        o[FTag.ClOrdID] = self.clord_id

        # setting instrument identification fields (may vary for different FIX brokers)
        self.set_instrument(o)

        # setting account field (may vary for different FIX brokers)
        self.set_account(o)
        o[FTag.OrdType] = self.ord_type
        o[FTag.Side] = self.side
        o[FTag.TransactTime] = self.current_datetime()

        # setting price and qty (typically tick size rounding, etc)
        self.set_price_qty(o, self.price, self.qty)

        return o

    def cancel_req(self) -> FIXMessage:
        """
        Creates order cancel request

        Raises:
            FIXError: if order can't be canceled

        """
        if not self.can_cancel():
            raise FIXError(f"{self} Fix order is not allowed for cancel")

        assert not self.orig_clord_id
        self.orig_clord_id = self.clord_id
        self.clord_id = self.next_clord()

        cxl_req_msg = FIXMessage(FMsg.ORDERCANCELREQUEST)
        cxl_req_msg[11] = self.clord_id
        cxl_req_msg[38] = self.qty
        cxl_req_msg[41] = self.clord_id
        self.set_instrument(cxl_req_msg)
        cxl_req_msg[FTag.Side] = self.side
        cxl_req_msg[FTag.TransactTime] = self.current_datetime()

        return cxl_req_msg

    def replace_req(self, price: float = nan, qty: float = nan) -> FIXMessage:
        """
        Creates order replace request

        Args:
            price: alternative price
            qty: alternative qty

        Returns:
            message

        Raises:
            FIXError: if order can't be replaced or price/qty unchanged

        """
        if not self.can_replace():
            raise FIXError("Order cannot be replaced")

        if not isfinite(price) or price == self.price:
            price = self.price
        if not isfinite(qty) or qty == self.qty or qty == 0:
            qty = self.qty

        if price == self.price and qty == self.qty:
            raise FIXError("no price / qty change in replace_req")

        assert not self.orig_clord_id
        self.orig_clord_id = self.clord_id
        self.clord_id = self.next_clord()

        m = FIXMessage(FMsg.ORDERCANCELREPLACEREQUEST)
        m[FTag.ClOrdID] = self.clord_id
        m[FTag.OrigClOrdID] = self.orig_clord_id
        m[FTag.OrdType] = self.ord_type
        self.set_instrument(m)
        self.set_price_qty(m, price, qty)
        m[FTag.Side] = self.side
        m[FTag.TransactTime] = self.current_datetime()

        return m

    def set_instrument(self, ord_msg: FIXMessage):
        """
        Set order instrument definition (override this in child)

        Args:
            ord_msg: new or replaced order

        """
        # Simply populate symbol, in read life FIX counterparty may require populate
        #  more fields (override this method in child class)
        ord_msg[FTag.Symbol] = self.ticker

    def set_account(self, ord_msg: FIXMessage):
        """
        Set account definition (override this in child)

        Args:
            ord_msg: new or replaced order

        """
        # Simplistic account setting (you must override this method in child class)
        assert isinstance(self.account, str), "Account expected a string"
        ord_msg[FTag.Account] = self.account

    def set_price_qty(self, ord_msg: FIXMessage, price: float, qty: float):
        """
        Set order price and qty definition (override this in child)

        This method handles custom price/qty rounding/decimal formatting, or maybe
        conditional presence of two fields based on order type

        Args:
            ord_msg: new or replaced order

        """
        ord_msg[FTag.Price] = price
        ord_msg[FTag.OrderQty] = qty

    @staticmethod
    def change_status(
        status: FOrdStatus,
        fix_msg_type: FMsg,
        msg_exec_type: FExecType,
        msg_status: FOrdStatus,
        raise_on_err: bool = True,
    ) -> FOrdStatus | None:
        """
        FIX Order State transition algo

        :param status: current order status
        :param fix_msg_type: incoming/or requesting order type,  these are supported:
                '8' - execution report,
                '9' - Order Cancel reject,
                'F' - Order cancel request (if possible to cancel current order)
                'G' -  Order replace request (if possible to replace current order)
        :param msg_exec_type: (only for execution report), for other should be 0
        :param msg_status: new fix msg order status, or required status
        :return: FOrdStatus if state transition is possible,
                 None - if transition is valid, but need to wait for a good state
                 raises FIXError - when transition is invalid
        """
        status_transitions = {}

        if fix_msg_type == FMsg.EXECUTIONREPORT:
            status_transitions = {
                None: {None: FIXError},
                # key {initial status}: {
                #    msg_status: <transition>,
                #       <transition>: None - ignore, True - transit, FIXError - raise
                #
                #      REJECTED: True,     #  this is allowed status transition
                #      PENDING_NEW: None,  #  transition allowed but no status change
                #      CREATED: FIXError,  #  error transition
                #      None:  [None, True, FIXError] # default transition
                #  }
                FOrdStatus.CREATED: {
                    FOrdStatus.PENDING_NEW: True,
                    FOrdStatus.REJECTED: True,
                    None: FIXError,
                },
                FOrdStatus.PENDING_NEW: {
                    FOrdStatus.REJECTED: True,
                    FOrdStatus.NEW: True,
                    FOrdStatus.FILLED: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.SUSPENDED: True,
                    None: FIXError,
                },
                FOrdStatus.NEW: {
                    FOrdStatus.NEW: None,
                    FOrdStatus.PENDING_NEW: FIXError,
                    FOrdStatus.CREATED: FIXError,
                    FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                    None: True,
                },
                FOrdStatus.FILLED: {
                    None: None,
                },
                FOrdStatus.CANCELED: {
                    None: None,
                },
                FOrdStatus.REJECTED: {
                    None: None,
                },
                FOrdStatus.EXPIRED: {
                    None: None,
                },
                FOrdStatus.SUSPENDED: {
                    FOrdStatus.NEW: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.SUSPENDED: None,
                    None: FIXError,
                },
                FOrdStatus.PARTIALLY_FILLED: {
                    FOrdStatus.FILLED: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.PENDING_REPLACE: True,
                    FOrdStatus.PENDING_CANCEL: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.EXPIRED: True,
                    FOrdStatus.SUSPENDED: True,
                    FOrdStatus.STOPPED: True,
                    None: FIXError,
                },
                FOrdStatus.PENDING_CANCEL: {
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.CREATED: FIXError,
                    None: None,
                },
                FOrdStatus.PENDING_REPLACE: {
                    "exec_type": {
                        FExecType.REPLACED: {
                            FOrdStatus.NEW: True,
                            FOrdStatus.PARTIALLY_FILLED: True,
                            FOrdStatus.FILLED: True,
                            FOrdStatus.CANCELED: True,
                            None: FIXError,
                        },
                        None: {
                            FOrdStatus.CREATED: FIXError,
                            FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                            None: None,
                        },
                    },
                },
            }

        elif fix_msg_type == FMsg.ORDERCANCELREJECT:  # '9'
            status_transitions = {
                None: {
                    FOrdStatus.CREATED: FIXError,
                    FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                    None: True,
                }
            }
        elif (
            fix_msg_type == FMsg.ORDERCANCELREQUEST
            or fix_msg_type == FMsg.ORDERCANCELREPLACEREQUEST
        ):
            status_transitions = {
                FOrdStatus.PENDING_CANCEL: {None: None},
                FOrdStatus.PENDING_REPLACE: {None: None},
                FOrdStatus.NEW: {None: True},
                FOrdStatus.SUSPENDED: {None: True},
                FOrdStatus.PARTIALLY_FILLED: {None: True},
                None: {None: FIXError},
            }

        if not status_transitions:
            raise FIXError(f"No status transition table for {fix_msg_type=}")

        s = status_transitions.get(status, status_transitions[None])
        if isinstance(s, dict) and "exec_type" in s:
            s = s["exec_type"].get(msg_exec_type, s["exec_type"][None])
        default = s[None]
        result = s.get(msg_status, default)

        if result is FIXError:
            # Invalid transition
            if raise_on_err:
                raise FIXError("FIX Order state transition error")
            else:
                return None
        else:
            if result is None:
                # Valid transition no status change
                return None
            else:
                # Do status change
                return msg_status

    def process_cancel_rej_report(self, m: FIXMessage) -> bool:
        """
        Processes incoming cancel reject report message
        """
        if m.msg_type != FMsg.ORDERCANCELREJECT:
            raise FIXError("incorrect message type")

        order_status = m[FTag.OrdStatus]

        new_status = self.change_status(
            self.status, m.msg_type, 0, order_status, raise_on_err=False
        )
        #
        if order_status == FOrdStatus.REJECTED:
            # Very weird (emergency) case, because the ClOrdId does not exist
            #   Let's set order inactive
            self.leaves_qty = 0

        if new_status is not None:
            self.status = new_status
            return True
        else:
            return False

    def process_execution_report(self, m: FIXMessage) -> bool:
        """
        Processes incoming execution report for an order

        Raises:
            FIXError: if ClOrdID mismatch
        """
        if m.msg_type != FMsg.EXECUTIONREPORT:
            raise FIXError("incorrect message type")

        clord_id = m[FTag.ClOrdID]
        cum_qty = float(m[FTag.CumQty])
        order_status = m[FTag.OrdStatus]

        if clord_id != self.clord_id and clord_id != self.orig_clord_id:
            raise FIXError("orig_clord_id mismatch")

        exec_type = m[FTag.ExecType]
        leaves_qty = float(m[FTag.LeavesQty])

        new_status = FIXNewOrderSingle.change_status(
            self.status, m.msg_type, exec_type, order_status, raise_on_err=False
        )

        self.order_id = m[FTag.OrderID]
        self.leaves_qty = leaves_qty
        self.cum_qty = cum_qty

        if exec_type == FExecType.REPLACED:
            # Order has been successfully replaced
            new_price = m.get(FTag.Price, None)
            if new_price is not None:
                # Price may not be in execution report, it's not an error
                self.price = float(new_price)

            order_qty = m.get(FTag.OrderQty, None)
            if order_qty is not None:
                # Qty may not be in execution report, it's not an error
                self.qty = float(order_qty)

            # Clearing OrigOrdId to allow subsequent order changes
            self.orig_clord_id = None

        if new_status:
            self.status = new_status
            return True
        else:
            return False

    def is_finished(self) -> bool:
        """
        Check if order is in terminal state (no subsequent changes expected)
        """
        return (
            self.status == FOrdStatus.FILLED
            or self.status == FOrdStatus.CANCELED
            or self.status == FOrdStatus.REJECTED
            or self.status == FOrdStatus.EXPIRED
        )

    def can_cancel(self) -> bool:
        """
        Check if order can be canceled from its current state
        """
        return (
            FIXNewOrderSingle.change_status(
                self.status,
                FMsg.ORDERCANCELREQUEST,
                0,  # Exec Type - omitted!
                FOrdStatus.PENDING_CANCEL,
                raise_on_err=False,
            )
            is not None
        )

    def can_replace(self) -> bool:
        """
        Check if order can be replaced from its current state
        """
        return (
            FIXNewOrderSingle.change_status(
                self.status,
                FMsg.ORDERCANCELREPLACEREQUEST,
                0,  # Exec Type - omitted!
                FOrdStatus.PENDING_REPLACE,
                raise_on_err=False,
            )
            is not None
        )



import enum


class MsgTypeEnumMeta(enum.EnumMeta):
    def __contains__(cls, item):
        if isinstance(item, enum.Enum):
            return isinstance(item, cls) and item._name_ in cls._member_map_
        else:
            return item in cls._value2member_map_


class FMsg(enum.Enum, metaclass=MsgTypeEnumMeta):
    def __str__(self):
        return str(self.value)

    def __eq__(self, o):
        return self.value == o

    def __hash__(self):
        return hash(self.value)

    HEARTBEAT = "0"
    TESTREQUEST = "1"
    RESENDREQUEST = "2"
    REJECT = "3"
    SEQUENCERESET = "4"
    LOGOUT = "5"
    IOI = "6"
    ADVERTISEMENT = "7"
    EXECUTIONREPORT = "8"
    ORDERCANCELREJECT = "9"
    QUOTESTATUSREQUEST = "a"
    LOGON = "A"
    DERIVATIVESECURITYLIST = "AA"
    NEWORDERMULTILEG = "AB"
    MULTILEGORDERCANCELREPLACE = "AC"
    TRADECAPTUREREPORTREQUEST = "AD"
    TRADECAPTUREREPORT = "AE"
    ORDERMASSSTATUSREQUEST = "AF"
    QUOTEREQUESTREJECT = "AG"
    RFQREQUEST = "AH"
    QUOTESTATUSREPORT = "AI"
    QUOTERESPONSE = "AJ"
    CONFIRMATION = "AK"
    POSITIONMAINTENANCEREQUEST = "AL"
    POSITIONMAINTENANCEREPORT = "AM"
    REQUESTFORPOSITIONS = "AN"
    REQUESTFORPOSITIONSACK = "AO"
    POSITIONREPORT = "AP"
    TRADECAPTUREREPORTREQUESTACK = "AQ"
    TRADECAPTUREREPORTACK = "AR"
    ALLOCATIONREPORT = "AS"
    ALLOCATIONREPORTACK = "AT"
    CONFIRMATIONACK = "AU"
    SETTLEMENTINSTRUCTIONREQUEST = "AV"
    ASSIGNMENTREPORT = "AW"
    COLLATERALREQUEST = "AX"
    COLLATERALASSIGNMENT = "AY"
    COLLATERALRESPONSE = "AZ"
    NEWS = "B"
    MASSQUOTEACKNOWLEDGEMENT = "b"
    COLLATERALREPORT = "BA"
    COLLATERALINQUIRY = "BB"
    NETWORKCOUNTERPARTYSYSTEMSTATUSREQUEST = "BC"
    NETWORKCOUNTERPARTYSYSTEMSTATUSRESPONSE = "BD"
    USERREQUEST = "BE"
    USERRESPONSE = "BF"
    COLLATERALINQUIRYACK = "BG"
    CONFIRMATIONREQUEST = "BH"
    EMAIL = "C"
    SECURITYDEFINITIONREQUEST = "c"
    SECURITYDEFINITION = "d"
    NEWORDERSINGLE = "D"
    SECURITYSTATUSREQUEST = "e"
    NEWORDERLIST = "E"
    ORDERCANCELREQUEST = "F"
    SECURITYSTATUS = "f"
    ORDERCANCELREPLACEREQUEST = "G"
    TRADINGSESSIONSTATUSREQUEST = "g"
    ORDERSTATUSREQUEST = "H"
    TRADINGSESSIONSTATUS = "h"
    MASSQUOTE = "i"
    BUSINESSMESSAGEREJECT = "j"
    ALLOCATIONINSTRUCTION = "J"
    BIDREQUEST = "k"
    LISTCANCELREQUEST = "K"
    BIDRESPONSE = "l"
    LISTEXECUTE = "L"
    LISTSTRIKEPRICE = "m"
    LISTSTATUSREQUEST = "M"
    XMLNONFIX = "n"
    LISTSTATUS = "N"
    REGISTRATIONINSTRUCTIONS = "o"
    REGISTRATIONINSTRUCTIONSRESPONSE = "p"
    ALLOCATIONINSTRUCTIONACK = "P"
    ORDERMASSCANCELREQUEST = "q"
    DONTKNOWTRADEDK = "Q"
    QUOTEREQUEST = "R"
    ORDERMASSCANCELREPORT = "r"
    QUOTE = "S"
    NEWORDERCROSS = "s"
    SETTLEMENTINSTRUCTIONS = "T"
    CROSSORDERCANCELREPLACEREQUEST = "t"
    CROSSORDERCANCELREQUEST = "u"
    MARKETDATAREQUEST = "V"
    SECURITYTYPEREQUEST = "v"
    SECURITYTYPES = "w"
    MARKETDATASNAPSHOTFULLREFRESH = "W"
    SECURITYLISTREQUEST = "x"
    MARKETDATAINCREMENTALREFRESH = "X"
    MARKETDATAREQUESTREJECT = "Y"
    SECURITYLIST = "y"
    QUOTECANCEL = "Z"
    DERIVATIVESECURITYLISTREQUEST = "z"



from __future__ import annotations

import dataclasses
import datetime as dtm
import re
import warnings
import xml.etree.ElementTree as ET
from math import isfinite

from asyncfix import FIXMessage, FTag
from asyncfix.errors import FIXMessageError
from asyncfix.message import FIXContainer


@dataclasses.dataclass
class SchemaField:
    tag: str
    name: str
    ftype: str
    values: dict[str, str] = dataclasses.field(default_factory=dict)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, o):
        if isinstance(o, SchemaField):
            return self.name == o.name
        elif isinstance(o, (str, int)):
            try:
                return self.tag == str(int(o))
            except Exception:
                return self.name == o
        else:
            return False

    def validate_value(self, value: str) -> bool:
        assert isinstance(value, str), "value must be a string"
        assert value, "empty value"

        if self.values:
            if value not in self.values:
                raise FIXMessageError(
                    f"Field={self.name} value expected to be one of the"
                    f" {list(self.values.keys())}, got ({value=})"
                )
            return True
        else:
            t = self.ftype.upper()
            err = None
            if t == "INT":
                err = SchemaField._validate_value_number(value, int)
            elif t in ["SEQNUM", "NUMINGROUP"]:
                err = SchemaField._validate_value_number(
                    value,
                    int,
                    no_zero=True,
                    no_negative=True,
                )
            elif t == "DAYOFMONTH":
                err = SchemaField._validate_value_number(value, int, num_range=(1, 31))
            elif t in {"FLOAT", "QTY", "PRICE", "PRICEOFFSET", "AMT", "PERCENTAGE"}:
                err = SchemaField._validate_value_number(
                    value, float, no_nonfinite=True
                )
            elif t in {"STRING", "MULTIPLESTRINGVALUE"}:
                err = SchemaField._validate_value_str(value)
            elif t in {"CHAR"}:
                err = SchemaField._validate_value_str(value, max_len=1)
            elif t in {"BOOLEAN"}:
                err = SchemaField._validate_value_str(
                    value, max_len=1, subset={"Y", "N"}
                )
            elif t in {"COUNTRY"}:
                err = SchemaField._validate_value_str(value, max_len=2, alpha_num=True)
            elif t in {"CURRENCY"}:
                err = SchemaField._validate_value_str(value, max_len=3, alpha_num=True)
            elif t in {"EXCHANGE"}:
                err = SchemaField._validate_value_str(value, max_len=4, alpha_num=True)
            elif t in {"LOCALMKTDATE", "UTCDATEONLY"}:
                err = SchemaField._validate_value_datetime(value, "%Y%m%d")
            elif t in {"UTCTIMESTAMP"}:
                err = SchemaField._validate_value_datetime(value, "%Y%m%d-%H:%M:%S")
            elif t in {"UTCTIMEONLY"}:
                err = SchemaField._validate_value_datetime(value, "%H:%M:%S")
            elif t == "MONTHYEAR":
                err = SchemaField._validate_value_monthyear(value)
            elif t in {"DATA", "LENGTH"}:
                # just hoping the data is ok
                err = None
            else:
                warnings.warn(f"Unsupported datatype: {t} for field={self}")

            if err:
                raise FIXMessageError(f"{self} validation error (value={value}): {err}")

            return True

    @staticmethod
    def _validate_value_datetime(value, format):
        if "." in value and "%S" in format and "%f" not in format:
            format = f"{format}.%f"

        try:
            dtm.datetime.strptime(value, format)
            return None  # all good
        except Exception as exc:
            return str(exc)

    @staticmethod
    def _validate_value_monthyear(value):
        """
        Special case for MonthYear type
            String representing month of a year. An optional day of the month can be
              appended or an optional week code.
            Valid formats:
            YYYYMM
            YYYYMMDD
            YYYYMMWW

            Valid values:
            YYYY = 0000-9999; MM = 01-12; DD = 01-31;
            WW = w1, w2, w3, w4, w5.
        """
        if "w" in value:
            week = value[-2:]
            if week not in {"w1", "w2", "w3", "w4", "w5"}:
                return f"YYYYMMWW [ww must be in w1,w2,w3,w4,w5], got {week}"
            format = "%Y%m"
            value = value[:-2]
            if len(value) != 6:
                return "remainder YYYYMM invalid"
        else:
            if len(value) == 6:
                format = "%Y%m"
            else:
                format = "%Y%m%d"

        return SchemaField._validate_value_datetime(value, format)

    @staticmethod
    def _validate_value_str(value, max_len=None, subset=None, alpha_num=False):
        assert value
        assert type(value) is str
        if "\x01" in value:
            return "Value contains SOH char"
        if "=" in value:
            return "Value contains `=` char"

        if max_len and len(value) > max_len:
            return "max legth exceeded"
        if subset and value not in subset:
            return f"out of subset: {subset}"
        if alpha_num and re.search(r"\W+", value):
            return "value contains non alphanumeric letters"

    @staticmethod
    def _validate_value_number(
        value: str,
        num_type: type,
        no_zero=False,
        no_negative=False,
        no_nonfinite=False,
        num_range=None,
    ) -> str | None:
        assert value
        assert num_type in (int, float)

        try:
            v = num_type(value)
            if no_zero and v == 0:
                raise ValueError("zero value")
            if no_negative and v < 0:
                raise ValueError("negative value")
            if no_nonfinite and not isfinite(float(v)):
                raise ValueError("not isfinite number")
            if num_range and not (v >= num_range[0] and v <= num_range[1]):
                raise ValueError(f"out of range {num_range}")
            # all good
            return None
        except ValueError as exc:
            return str(exc)

    def __str__(self):
        return f"{self.name}|{self.tag}"

    def __repr__(self):
        return f"SchemaField({str(self)}, type={self.ftype})"


class SchemaSet:
    def __init__(self, name: str, field: SchemaField | None = None):
        self.name: str = name
        self.field: SchemaField = field  # this is set for groups, with attached field
        if field:
            if not (
                ("No" in field.name or "Num" in field.name)
                and field.ftype in ["NUMINGROUP", "INT"]
            ):
                # warnings.warn('')
                raise ValueError(
                    "SchemaSet expected to have group field with NUMINGROUP|INT type,"
                    f" name contais No/Num, got {field}"
                )
        self.members: dict[SchemaField | SchemaSet, SchemaField | SchemaSet] = {}
        self.required: dict[SchemaField | SchemaSet, bool] = {}

    @property
    def tag(self) -> str:
        if self.field:
            return self.field.tag
        else:
            raise ValueError(f"tag property is not supported for {self}")

    def keys(self) -> list[str]:
        return [m.name for m in self.members]

    def add(self, field_or_set: SchemaField | SchemaSet, required: bool):
        if isinstance(field_or_set, SchemaField):
            assert field_or_set not in self.members
        elif isinstance(field_or_set, SchemaSet):
            assert field_or_set.field, "field_or_set.field is empty, try to merge"
        else:
            raise ValueError(f"Unsupported field_or_set type, got {type(field_or_set)}")

        self.members[field_or_set] = field_or_set
        self.required[field_or_set] = required

    def merge(self, comp: SchemaSet):
        assert isinstance(comp, SchemaSet)

        for field, required in comp.members.items():
            self.add(field, comp.required[field])

    def __hash__(self):
        if self.field:
            return hash(self.field)
        else:
            return hash(self.name)

    def __eq__(self, o):
        if self.field:
            return self.field == o
        else:
            return self.name == o

    def __contains__(self, item: str | SchemaField) -> bool:
        try:
            int(str(item))
            raise FIXMessageError("item looks like tag, use name or SchemaField")
        except ValueError:
            return item in self.members

    def __getitem__(self, item: str | SchemaField) -> SchemaField | SchemaSet:
        try:
            int(str(item))
            raise FIXMessageError("item looks like tag, use name or SchemaField")
        except ValueError:
            return self.members[item]


class SchemaGroup(SchemaSet):
    def __init__(self, field: SchemaField, required: bool):
        super().__init__(field.name, field)
        self.field_required = required

    def validate_group(self, groups: list[FIXContainer]):
        tag_order = {f.tag: i for i, f in enumerate(self.members.values())}
        tag_fields = {f.tag: f for f in self.members.values()}

        for fmsg in groups:
            has_first_tag = False
            prev_tag = -1

            for t, v in fmsg.items():
                if t not in tag_order:
                    raise FIXMessageError(
                        f"fixmessage={groups} contains unsupported tag for {self}"
                    )
                ord_idx = tag_order[t]

                if ord_idx == 0:
                    has_first_tag = True
                if prev_tag > ord_idx:
                    raise FIXMessageError(
                        f"fixmessage={groups} incorrect tag order {self}"
                    )

                field = tag_fields[t]

                if isinstance(field, SchemaField):
                    field.validate_value(v)
                else:
                    # Nested group!?
                    assert isinstance(field, SchemaGroup)
                    if not fmsg.is_group(t):
                        raise FIXMessageError(
                            f"fixmessage={groups}, tag={t} must be a group {self}"
                        )

                    # validate nested group
                    field.validate_group(fmsg.get_group_list(t))

                prev_tag = ord_idx

            if not has_first_tag:
                raise FIXMessageError(
                    f"fixmessage={groups} does not contain mandatory first tag {self}"
                )

            for st, sv in tag_fields.items():
                if isinstance(sv, SchemaField):
                    if sv.tag not in fmsg and self.required[sv]:
                        raise FIXMessageError(
                            f"fixmessage={groups} missing required field {repr(sv)}"
                        )

    def __repr__(self):
        members = [str(m) for m in self.members.keys()]
        return f"SchemaGroup({self.field.name}, {members})"


class SchemaComponent(SchemaSet):
    def __init__(self, name: str):
        super().__init__(name)


class SchemaMessage(SchemaSet):
    def __init__(self, name: str, msg_type: str, msg_cat: str):
        super().__init__(name)
        self.msg_type = msg_type
        self.msg_cat = msg_cat

    def __repr__(self):
        return (
            f"SchemaMessage(name={self.name}, type={self.msg_type}, cat={self.msg_cat})"
        )


class FIXSchema:
    def __init__(self, xml: ET.ElementTree):
        assert isinstance(xml, ET.ElementTree)
        self.tag2field: dict[str, SchemaField] = {}
        self.field2tag: dict[str, SchemaField] = {}
        self.components: dict[str, SchemaComponent] = {}
        self.messages: dict[str, SchemaMessage] = {}
        self.messages_types: dict[str, SchemaMessage] = {}
        self.header = {}
        self.types = set()

        self._parse(xml.getroot())

    def _parse_msg_set(self, component, element):
        has_circular_refs = False
        for val in element:
            assert val.tag in ["field", "group", "component"]

            if val.tag == "field":
                field = self.field2tag[val.attrib["name"]]
                component.add(field, val.attrib["required"].upper() == "Y")
            elif val.tag == "component":
                if val.attrib["name"] not in self.components:
                    # We have circular reference, component was referenced before
                    #    described in XML data, just skip and rerun later
                    has_circular_refs = True
                    continue
                ref_component = self.components[val.attrib["name"]]
                component.merge(ref_component)
            elif val.tag == "group":
                g = self._parse_group(val)
                if g is None:
                    # Group also refers to other component, postpone it
                    has_circular_refs = True
                    continue
                component.add(g, g.required)

        if has_circular_refs:
            return None
        else:
            return component

    def _parse_group(self, element: ET.Element) -> SchemaGroup | None:
        assert self.field2tag, "parse fields first!"
        assert element.tag == "group"

        el_name = element.attrib["name"]

        gfield = self.field2tag[el_name]
        group = self._parse_msg_set(
            SchemaGroup(gfield, element.attrib["required"].upper() == "Y"),
            element,
        )
        # assert group, f'group parse erorr, {element.attrib}'
        return group

    def _parse_component(self, element: ET.Element) -> SchemaComponent | None:
        assert self.field2tag, "parse fields first!"
        assert element.tag == "component"

        el_name = element.attrib["name"]

        assert el_name not in self.components, "Duplicate component name or double run"
        component = self._parse_msg_set(SchemaComponent(el_name), element)

        if component:
            self.components[el_name] = component
            return component
        else:
            return None

    def _parse_message(self, element: ET.Element) -> SchemaMessage | None:
        assert self.field2tag, "parse fields first!"
        assert element.tag == "message"

        el_name = element.attrib["name"]

        assert el_name not in self.messages, "Duplicate message?"

        message = self._parse_msg_set(
            SchemaMessage(
                el_name,
                msg_type=element.attrib["msgtype"],
                msg_cat=element.attrib["msgcat"],
            ),
            element,
        )
        assert message, "Message probably refers to circular refs in comp or groups"

        self.messages[el_name] = message
        self.messages_types[message.msg_type] = message
        return message

    def _parse_field(self, element: ET.Element):
        assert element.tag == "field"

        f = SchemaField(
            tag=element.attrib["number"],
            name=element.attrib["name"],
            ftype=element.attrib["type"],
        )
        if len(element) > 0:
            for val in element:
                assert val.tag == "value"
                f.values[val.attrib["enum"]] = val.attrib["description"]

        self.types.add(f.ftype)
        self.tag2field[f.tag] = f
        self.field2tag[f.name] = f

    def _parse(self, root: ET.Element):
        assert not self.field2tag, "already parsed"

        for element in root.find("fields"):
            self._parse_field(element)

        all_components = [e for e in root.find("components")]
        full_count = len(all_components)
        prev_cnt = len(all_components)
        while all_components:
            i = 0
            while i < len(all_components):
                if self._parse_component(all_components[i]):
                    # component was parsed
                    del all_components[i]
                else:
                    i += 1

            if all_components:
                if len(all_components) == prev_cnt:
                    failed_comp = [c.attrib["name"] for c in all_components]
                    raise RuntimeError(
                        "Failed to resolve component circular reference."
                        f"Failed components: {failed_comp}"
                    )
                else:
                    prev_cnt = len(all_components)
        assert full_count == len(self.components), "Component count mismatch"

        n_msg = 0
        for element in root.find("messages"):
            self._parse_message(element)
            n_msg += 1

        assert n_msg == len(self.messages), "Message count mismatch"

    def validate(self, msg: FIXMessage) -> bool:
        if msg.msg_type not in self.messages_types:
            raise FIXMessageError(f"msg_type=`{msg.msg_type}` not in schema")

        schema_msg = self.messages_types[msg.msg_type]

        schema_fields = set()
        for fname, req in schema_msg.required.items():
            f = schema_msg[fname]
            schema_fields.add(fname)

            if req:
                if isinstance(f, SchemaField):
                    if f.tag not in msg:
                        raise FIXMessageError(f"Missing required field={repr(f)}")

        for tag, val in msg.tags.items():
            if tag not in self.tag2field:
                raise FIXMessageError(f"msg tag={tag} not in schema")
            field = self.tag2field[tag]
            if field not in schema_msg:
                raise FIXMessageError(
                    f"msg field={field} is not allowed in {schema_msg}"
                )

            fschema = schema_msg[field]
            if isinstance(fschema, SchemaField):
                if msg.is_group(tag):
                    raise FIXMessageError(
                        f"msg tag={tag} val={val} must be a tag, got group"
                    )
                fschema.validate_value(val)

            elif isinstance(fschema, SchemaGroup):
                if not msg.is_group(tag):
                    raise FIXMessageError(f"msg tag={tag} val={val} must be a group")
                fschema.validate_group(msg.get_group_list(tag))

        return True

    def __getitem__(self, item: int | str | FTag) -> SchemaField:
        try:
            tag = str(int(str(item)))
            return self.tag2field[tag]
        except ValueError:
            return self.field2tag[str(item)]



import logging
from datetime import datetime

from asyncfix import FTag
from asyncfix.errors import EncodingError
from asyncfix.message import FIXContainer, FIXMessage, RepeatingTagError
from asyncfix.protocol import FIXProtocolBase
from asyncfix.session import FIXSession


class RepeatingGroupContext(FIXContainer):
    def __init__(self, tag, repeating_group_tags, parent):
        self.tag = tag
        self.repeating_group_tags = repeating_group_tags
        self.parent = parent
        FIXContainer.__init__(self)


class Codec(object):
    def __init__(self, protocol: FIXProtocolBase):
        self.protocol: FIXProtocolBase = protocol
        self.SOH = "\x01"

    @staticmethod
    def current_datetime():
        return datetime.utcnow().strftime("%Y%m%d-%H:%M:%S.%f")[:-3]

    def _addTag(self, body, t, msg: FIXMessage):
        if msg.is_group(t):
            groups = msg.get_group_list(t)
            body.append("%s=%s" % (t, len(groups)))
            for group in groups:
                for tag in group.tags:
                    self._addTag(body, tag, group)
        else:
            body.append("%s=%s" % (t, msg[t]))

    def encode(self, msg: FIXMessage, session: FIXSession) -> str:
        # Create body
        body = []

        msg_type = msg.msg_type

        body.append(
            "%s=%s" % (self.protocol.fixtags.SenderCompID, session.sender_comp_id)
        )
        body.append(
            "%s=%s" % (self.protocol.fixtags.TargetCompID, session.target_comp_id)
        )

        seq_no = 0
        if msg_type == self.protocol.msgtype.SEQUENCERESET:
            if (
                self.protocol.fixtags.GapFillFlag in msg
                and msg[self.protocol.fixtags.GapFillFlag] == "Y"
            ):
                # in this case the sequence number should already be on the message
                try:
                    seq_no = msg[self.protocol.fixtags.MsgSeqNum]
                except KeyError:
                    raise EncodingError(
                        "SequenceReset with GapFill='Y' must have the MsgSeqNum already"
                        " populated"
                    )
            else:
                msg[self.protocol.fixtags.NewSeqNo] = session.allocate_snd_seq_no()
                seq_no = msg[self.protocol.fixtags.MsgSeqNum]
        else:
            # if we have the PossDupFlag set, we need to send the message
            #   with the same seqNo
            if (
                self.protocol.fixtags.PossDupFlag in msg
                and msg[self.protocol.fixtags.PossDupFlag] == "Y"
            ):
                try:
                    seq_no = msg[self.protocol.fixtags.MsgSeqNum]
                except KeyError:
                    raise EncodingError(
                        "Failed to encode message with PossDupFlay=Y but no previous"
                        " MsgSeqNum"
                    )
            else:
                seq_no = session.allocate_snd_seq_no()

        body.append("%s=%s" % (self.protocol.fixtags.MsgSeqNum, seq_no))
        body.append(
            "%s=%s" % (self.protocol.fixtags.SendingTime, self.current_datetime())
        )

        for t in msg.tags:
            self._addTag(body, t, msg)

        # Enable easy change when debugging
        SEP = self.SOH

        body = self.SOH.join(body) + self.SOH

        # Create header
        header = []
        msg_type = "%s=%s" % (self.protocol.fixtags.MsgType, msg_type)
        header.append(
            "%s=%s" % (self.protocol.fixtags.BeginString, self.protocol.beginstring)
        )
        header.append(
            "%s=%i" % (self.protocol.fixtags.BodyLength, len(body) + len(msg_type) + 1)
        )
        header.append(msg_type)

        fixmsg = self.SOH.join(header) + self.SOH + body

        cksum = sum([ord(i) for i in fixmsg]) % 256
        fixmsg = fixmsg + "%s=%0.3i" % (self.protocol.fixtags.CheckSum, cksum)

        # print len(fixmsg)

        return fixmsg + SEP

    def decode(
        self, rawmsg: bytes, silent: bool = True
    ) -> tuple[FIXMessage | None, int]:
        valid_idx = rawmsg.find(b"8=FIX.")
        if valid_idx == -1:
            assert silent, "no fix header"
            return None, len(rawmsg)

        parsed_length = valid_idx

        msg = rawmsg[valid_idx:].decode("latin-1")

        next_msg = msg[5:].find("8=FIX.")
        if next_msg != -1:
            # Next fix message added, but incomplete
            next_msg += 5
        else:
            next_msg = len(msg)

        msg = msg[:next_msg].split(self.SOH)
        if not msg[-1]:
            msg = msg[:-1]

        # at a minimum we require BeginString, BodyLength & Checksum
        if len(msg) < 3:
            assert silent, "Minimum message"
            return (None, parsed_length)

        tag, value = msg[0].split("=", 1)
        if value != self.protocol.beginstring:
            logging.error(
                "FIX Version unexpected (Recv: %s Expected: %s)"
                % (value, self.protocol.beginstring)
            )
            assert silent, "protocol beginstring mismatch"
            return (None, len(rawmsg))

        toks = msg[1].split("=", 1)
        if len(toks) != 2:
            assert silent, f"BodyLength split error {msg}"
            return (None, len(rawmsg))
        tag, value = toks

        msg_length = len(msg[0]) + len(msg[1]) + len("10=000") + 3
        if tag != FTag.BodyLength:
            logging.error(f"*** BodyLength missing or not 2nd field *** [{tag}]: {msg}")
            assert silent, "2nd tag must be BodyLength"
            return (None, len(rawmsg))
        else:
            msg_length += int(value)

        # message looks incomplete
        if msg_length > len(rawmsg):
            assert silent, "incomplete message"
            return (None, parsed_length)

        checksum_passed = False
        parsed_length += msg_length

        decoded_msg = FIXMessage("UNKNOWN")
        repeating_groups = []
        repeating_group_tags = self.protocol.repeating_groups
        current_context = decoded_msg

        for m in msg:
            toks = m.split("=", 1)
            if len(toks) != 2:
                assert silent, f"incomplete tag {m}"
                return (None, len(rawmsg))
            tag, value = toks

            if tag == FTag.CheckSum:
                cheksum_base = self.SOH.join(msg[:-1])
                checksum = (sum([ord(i) for i in cheksum_base]) + 1) % 256

                if checksum != int(value):
                    logging.warning(
                        "\tCheckSum: %s (INVALID) expecting %s" % (int(value), checksum)
                    )
                    assert (
                        silent
                    ), f"invalid checksum tag[10]={value} expected: {checksum} {msg=}"
                    checksum_passed = False
                else:
                    checksum_passed = True
            elif tag == FTag.MsgType:
                decoded_msg.msg_type = value

            # found the start of a repeating group
            if tag in repeating_group_tags:
                # i.e. we are already in a repeating group
                if type(current_context) is RepeatingGroupContext:
                    while (
                        repeating_groups
                        and tag not in current_context.repeating_group_tags
                    ):
                        current_context.parent.add_group(
                            current_context.tag, current_context
                        )
                        current_context = current_context.parent
                        # pop the completed group off the stack
                        del repeating_groups[-1]

                ctx = RepeatingGroupContext(
                    tag, repeating_group_tags[tag], current_context
                )
                repeating_groups.append(ctx)
                current_context = ctx
            elif repeating_groups:
                # we have 1 or more repeating groups in progress
                #    & our tag isn't the start of a group
                while (
                    repeating_groups and tag not in current_context.repeating_group_tags
                ):
                    current_context.parent.add_group(
                        current_context.tag, current_context
                    )
                    current_context = current_context.parent
                    # pop the completed group off the stack
                    del repeating_groups[-1]

                if tag in current_context.tags:
                    # if the repeating group already contains this field,
                    #     start the next
                    current_context.parent.add_group(
                        current_context.tag, current_context
                    )
                    ctx = RepeatingGroupContext(
                        current_context.tag,
                        current_context.repeating_group_tags,
                        current_context.parent,
                    )
                    del repeating_groups[-1]
                    repeating_groups.append(ctx)
                    current_context = ctx

                # else add it to the current one
                current_context.set(tag, value)
            else:
                if tag in decoded_msg:
                    # Repeating tag found, possibly RepGrp not in protocol schema
                    decoded_msg.set(tag, RepeatingTagError)
                else:
                    # this isn't a repeating group field, so just add it normally
                    decoded_msg.set(tag, value)

        if checksum_passed:
            return (decoded_msg, parsed_length)
        else:
            assert silent, f"Checksum probably missing: {msg}"
            return (None, parsed_length)
